@@ -871,7 +871,7 @@ g_get_any_init (void)
 	gchar *home = g_getenv ("HOME");
       
 	/* Only believe HOME if it is an absolute path and exists */
-	if (g_path_is_absolute (home) && g_file_test (home, G_FILE_TEST_IS_DIR))
+	if (home && g_path_is_absolute (home) && g_file_test (home, G_FILE_TEST_IS_DIR))
 	  g_home_dir = g_strdup (home);
       }
       
@@ -934,7 +934,10 @@ g_get_any_init (void)
         do
           {
             g_free (buffer);
-            buffer = g_malloc (bufsize);
+	    /* we allocate 6 extra bytes to work around a bug in 
+	     * Mac OS < 10.3. See #156446
+             */
+            buffer = g_malloc (bufsize + 6);
 	    errno = 0;
 	    
 #    ifdef HAVE_POSIX_GETPWUID_R
@@ -1236,11 +1239,11 @@ g_nullify_pointer (gpointer *nullify_location)
 gchar *
 g_get_codeset (void)
 {
-  gchar *charset;
+  const gchar *charset;
 
   g_get_charset (&charset);
 
-  return charset;
+  return strdup (charset);
 }
 
 #ifdef ENABLE_NLS
